// Copyright (c) 2023 VMware Carbon Black, all rights reserved

package controllers

import (
	"errors"
	"fmt"
	"net/http"
	"net/http/httptest"
	"testing"
	"time"

	"gitlab.bit9.local/psc/ems/hinting/pkg/hints"
	"gitlab.bit9.local/psc/ems/hinting/pkg/hints/hinttype"
	"hello-api/internal/hinting"

	"hello-api/internal/app/middleware"
	"hello-api/internal/app/parsers/deviceresponseheader"

	"github.com/gin-gonic/gin"
	. "github.com/onsi/ginkgo/v2"
	. "github.com/onsi/gomega"
)

type errorEncoder struct{}

func (er *errorEncoder) Encode(_ *deviceresponseheader.DeviceResponseHeader) (string, error) {
	return "", errors.New("error")
}

type mockRouterGroup struct {
	handlers []gin.HandlerFunc
	route    string
}

func (mrg *mockRouterGroup) POST(prefix string, handlers ...gin.HandlerFunc) gin.IRoutes {
	mrg.handlers = handlers
	mrg.route = prefix
	return nil
}

var _ = Describe("hello controller", func() {
	helloController := NewHelloController()
	It("returns 200 code on valid inputs", func() {
		recorder := httptest.NewRecorder()
		testContext, _ := gin.CreateTestContext(recorder)
		hintResult := hints.NewHintResult(hinttype.NEW_CONFIG_DATA, hinttype.NEW_POLICY, hinttype.LIVE_QUERY)
		version := 18
		lastUpdateTime := time.UnixMilli(1675397106000)    // Friday, February 3, 2023 4:05:06
		lastLiveQueryTime := time.UnixMilli(1672628645000) // Monday, January 2, 2023 3:04:05
		hintResult.LastUpdateTime = &lastUpdateTime
		hintResult.LastLiveQueryTime = &lastLiveQueryTime
		testContext.Set(middleware.HintsName, hintResult)
		testContext.Set(middleware.VersionName, version)
		helloController.DoHello(testContext)
		Expect(recorder.Code).To(BeComparableTo(http.StatusOK))

		decoder := deviceresponseheader.NewResponseHeaderDecoder()
		responseBodyParsed, err := decoder.Decode(recorder.Body.String())
		Expect(err).To(BeNil())
		Expect(responseBodyParsed).To(BeEquivalentTo(deviceresponseheader.NewDeviceResponseHeader(int64(version), HelloResponseCode, true, 0, "Success.", hintResult.Hints.ToIntArray(), &lastUpdateTime, &lastLiveQueryTime)))
	})

	It("returns 500 when encoding the response fails", func() {
		helloController := helloController.WithEncoder(&errorEncoder{})
		recorder := httptest.NewRecorder()
		testContext, _ := gin.CreateTestContext(recorder)
		testContext.Request, _ = http.NewRequest("GET", "/hello", http.NoBody)
		helloController.DoHello(testContext)
		Expect(recorder.Code).To(BeComparableTo(http.StatusInternalServerError))
	})

	It("returns 500 when LastQueryTime result is missing", func() {
		helloController := helloController.WithEncoder(&errorEncoder{})
		recorder := httptest.NewRecorder()
		testContext, _ := gin.CreateTestContext(recorder)
		hints := hints.NewHintResult(hinttype.NEW_CONFIG_DATA, hinttype.NEW_POLICY, hinttype.LIVE_QUERY)
		testContext.Set(middleware.HintsName, hints)
		lastUpdateTime := time.UnixMilli(1675397106000) // Friday, February 3, 2023 4:05:06
		testContext.Set("LastUpdateTime", &lastUpdateTime)

		testContext.Request, _ = http.NewRequest("GET", "/hello", http.NoBody)
		helloController.DoHello(testContext)
		Expect(recorder.Code).To(BeComparableTo(http.StatusInternalServerError))
	})

	It("returns 500 when LastUpdateTime result is missing", func() {
		helloController := helloController.WithEncoder(&errorEncoder{})
		recorder := httptest.NewRecorder()
		testContext, _ := gin.CreateTestContext(recorder)
		hints := hints.NewHintResult(hinttype.NEW_CONFIG_DATA, hinttype.NEW_POLICY, hinttype.LIVE_QUERY)
		testContext.Set(middleware.HintsName, hints)
		lastLiveQueryTime := time.UnixMilli(1672628645000) // Monday, January 2, 2023 3:04:05
		testContext.Set("LastQueryTime", &lastLiveQueryTime)

		testContext.Request, _ = http.NewRequest("GET", "/hello", http.NoBody)
		helloController.DoHello(testContext)
		Expect(recorder.Code).To(BeComparableTo(http.StatusInternalServerError))
	})

	It("adds route with appropriate middleware", func() {
		helloController := NewHelloController()
		doHello := gin.HandlerFunc(helloController.DoHello)
		deviceContextMiddleware := gin.HandlerFunc(middleware.DeviceContextMiddleware)
		deviceRequestMiddleware := gin.HandlerFunc(middleware.DeviceRequestHeaderMiddleware)
		hintsClient := hinting.NewHintsClient(nil, "")
		hintingMiddleware := middleware.HintingMiddleware(hintsClient)
		mockRouter := mockRouterGroup{}
		lctMiddleware := middleware.LctPublisherMiddleware(nil)
		helloController.AddToRouter(&mockRouter, lctMiddleware, hintingMiddleware)
		//goland:noinspection GoPrintFunctions
		Expect(fmt.Sprintf("%v", mockRouter.handlers[len(mockRouter.handlers)-1])).To(BeIdenticalTo(fmt.Sprintf("%v", doHello)))
		//goland:noinspection GoPrintFunctions
		Expect(fmt.Sprintf("%v", mockRouter.handlers[len(mockRouter.handlers)-2])).To(BeIdenticalTo(fmt.Sprintf("%v", hintingMiddleware)))
		//goland:noinspection GoPrintFunctions
		Expect(fmt.Sprintf("%v", mockRouter.handlers[len(mockRouter.handlers)-3])).To(BeIdenticalTo(fmt.Sprintf("%v", lctMiddleware)))
		//goland:noinspection GoPrintFunctions
		Expect(fmt.Sprintf("%v", mockRouter.handlers[len(mockRouter.handlers)-4])).To(BeIdenticalTo(fmt.Sprintf("%v", deviceRequestMiddleware)))
		//goland:noinspection GoPrintFunctions
		Expect(fmt.Sprintf("%v", mockRouter.handlers[len(mockRouter.handlers)-5])).To(BeIdenticalTo(fmt.Sprintf("%v", deviceContextMiddleware)))
	})
})

func TestController(t *testing.T) {
	RegisterFailHandler(Fail)
	RunSpecs(t, "Hello controller tests")
}
