// Copyright (c) 2023 VMware Carbon Black, all rights reserved

package middleware

import (
	"net/http"
	"strings"
	"time"

	"hello-api/internal/app/parsers/deviceresponseheader"
	"hello-api/internal/logger"
	"hello-api/internal/utils"

	"carbonblack.com/protocols/psc/ems/server/deviceauth"
	"github.com/gin-gonic/gin"
	"github.com/google/uuid"
	"github.com/opentracing-contrib/go-gin/ginhttp"
	"github.com/opentracing/opentracing-go"
	"github.com/opentracing/opentracing-go/ext"
	"go.uber.org/zap"
)

// Tracing -- Handles middleware tracing
func (mwf *MiddlewareUtils) Tracing() gin.HandlerFunc {
	return ginhttp.Middleware(mwf.metricsClient.Tracer(), mwf.tracingSpanObserverOption(), mwf.tracingOperationNameOption())
}

func (mwf *MiddlewareUtils) GetTracingChain() gin.HandlersChain {
	if mwf.metricsClient != nil && mwf.metricsClient.Tracer() != nil {
		return gin.HandlersChain{
			mwf.Tracing(),
			mwf.GetSpanErrorTaggingMiddleware,
			RecoveryMiddleware,
			mwf.Logging,
			mwf.TracingRequestIdAndLogger,
			TimeoutMiddleware(mwf.cfg.HttpHandleTimeout),
		}
	}
	return gin.HandlersChain{
		RecoveryMiddleware,
		mwf.Logging,
		mwf.TracingRequestIdAndLogger,
		TimeoutMiddleware(mwf.cfg.HttpHandleTimeout),
	}
}

func (mwf *MiddlewareUtils) Logging(c *gin.Context) {
	defer func(start time.Time) {
		latency := time.Since(start)
		deviceId := int64(0)
		orgKey := "MISSING"
		deviceContext, problems := utils.ContextGet[*deviceauth.Context](c, DeviceContextHeader)
		if problems == nil {
			deviceId = deviceContext.Device.Id
			orgKey = deviceContext.Organization.OrgKey
		} else {
			deviceContext = &deviceauth.Context{}
		}
		version := c.GetInt(VersionName)
		lastUpdateTime, _ := utils.ContextGet[*time.Time](c, lastUpdateTimeName)
		lastLiveQueryTime, _ := utils.ContextGet[*time.Time](c, lastQueryTimeName)

		responseBackToSensor, err := utils.ContextGet[*deviceresponseheader.DeviceResponseHeader](c, "response")

		if err != nil || responseBackToSensor == nil {
			responseBackToSensor = &deviceresponseheader.DeviceResponseHeader{Message: "!!!DEVICE RESPONSE HEADER NOT FOUND!!!"}
		}

		// Fill in additional information seen in the transaction
		logFields := []zap.Field{
			zap.String("remote-address", c.Request.RemoteAddr),
			zap.String("method", c.Request.Method),
			zap.String("request-uri", c.Request.RequestURI),
			zap.String("accept", strings.Join(c.Request.Header["Accept"], " ")),
			zap.String("proto", c.Request.Proto),
			zap.Int("status", c.Writer.Status()),
			zap.Int("length", int(c.Request.ContentLength)),
			zap.String("user-agent", c.Request.UserAgent()),
			zap.Duration("latency", latency),
			zap.Strings("errors", c.Errors.Errors()),
			zap.Int64("device-id", deviceId),
			zap.String("org_key", orgKey),
			zap.Int("version", version),
			zap.Timep("request_last_update_time", lastUpdateTime),
			zap.Timep("request_last_livequery_time", lastLiveQueryTime),
			zap.Object("response_header", responseBackToSensor),
			zap.Any("device-context", deviceContext),
		}

		if c.Writer.Status() == http.StatusInternalServerError {
			logger.RequestLoggerFromContext(c.Request.Context()).Error("Server Error", logFields...)
		} else {
			logger.RequestLoggerFromContext(c.Request.Context()).Debug("", logFields...)
		}
	}(time.Now())

	c.Next()
}

func (mwf *MiddlewareUtils) TracingRequestIdAndLogger(c *gin.Context) {
	var err error
	if sp := opentracing.SpanFromContext(c.Request.Context()); sp != nil {
		err = opentracing.GlobalTracer().Inject(
			sp.Context(),
			opentracing.HTTPHeaders,
			opentracing.HTTPHeadersCarrier(c.Writer.Header()),
		)
	}
	requestID := c.Writer.Header().Get(mwf.cfg.RequestIDHeader)
	if requestID == "" {
		requestID = uuid.New().String() // New request?  Need an ID
	}
	c.Writer.Header().Set(mwf.cfg.RequestIDHeader, requestID)
	l := mwf.RequestLogger.With(zap.String("request-id", requestID))
	if err != nil {
		l.Info(err.Error())
	}
	*c.Request = *c.Request.Clone(logger.ContextWithRequestLogger(c.Request.Context(), l))
	c.Next()
}

func (mwf *MiddlewareUtils) tracingOperationNameOption() ginhttp.MWOption {
	return ginhttp.OperationNameFunc(func(r *http.Request) string {
		return r.RequestURI
	})
}

func (mwf *MiddlewareUtils) tracingSpanObserverOption() ginhttp.MWOption {
	return ginhttp.MWSpanObserver(func(sp opentracing.Span, r *http.Request) {
		sp.SetTag("http.remote_address", r.RemoteAddr)
		sp.SetTag("http.user_agent", r.UserAgent())
		sp.SetTag("http.accept", r.Header.Get("Accept"))
	})
}

func (mwf *MiddlewareUtils) GetSpanErrorTaggingMiddleware(c *gin.Context) {
	defer func(c *gin.Context) {
		if sp := opentracing.SpanFromContext(c.Request.Context()); sp != nil {
			wasErrorReported := c.GetBool("error") || len(c.Errors) > 0
			sp.SetTag(string(ext.Error), wasErrorReported)
		}
	}(c)
	c.Next()
}
