// Copyright (c) 2023 VMware Carbon Black, all rights reserved

// Package devicecontext handles the parsing of Device Context information stored in the request header
package devicecontext

import (
	"encoding/base64"
	"errors"
	"fmt"
	"strings"

	"carbonblack.com/protocols/psc/ems/server/deviceauth"

	"google.golang.org/protobuf/proto"
)

type (
	DeviceAuthResponseParser interface {
		ParseResponse(deviceContextHeader string) (*deviceauth.Context, error)
	}

	deviceAuthResponseParser struct{}
)

func NewDeviceAuthResponseParser() DeviceAuthResponseParser {
	return &deviceAuthResponseParser{}
}

// ParseResponse -- Decode the protobuf response returned as a Base64 encoded string
//   - as per 'device_auth.proto'
func (parser *deviceAuthResponseParser) ParseResponse(deviceContextHeader string) (context *deviceauth.Context, err error) {
	if strings.TrimSpace(deviceContextHeader) == "" {
		return nil, errors.New("device context header was empty")
	}

	decodedHeader, err := base64.StdEncoding.DecodeString(deviceContextHeader)
	if err != nil {
		return nil, fmt.Errorf("failed to b64 decode device context header %w", err)
	}

	context = new(deviceauth.Context)
	err = proto.Unmarshal(decodedHeader, context)
	if err != nil {
		return nil, fmt.Errorf("failed to unmarshal device context header %w", err)
	}
	err = validateContext(context)
	if err != nil {
		return nil, err
	} else {
		return context, nil
	}
}

func validateContext(context *deviceauth.Context) (err error) {
	if context.Device == nil {
		return errors.New("no device included in device context header")
	}
	if context.DeviceGroup == nil {
		return errors.New("no device group included in device context header")
	}
	if context.Organization == nil {
		return errors.New("no org included in device context header")
	}
	return nil
}
