// Copyright (c) 2023 VMware Carbon Black, all rights reserved

package hinting

import (
	"bytes"
	"errors"
	"io"
	"net/http"
	"testing"
	"time"

	"gitlab.bit9.local/psc/ems/hinting/pkg/hints"
	"gitlab.bit9.local/psc/ems/hinting/pkg/hints/hinttype"

	. "github.com/onsi/ginkgo/v2"
	. "github.com/onsi/gomega"
	"github.com/pact-foundation/pact-go/v2/consumer"
	"github.com/pact-foundation/pact-go/v2/matchers"
)

var t *testing.T

func TestHintingClient(test *testing.T) {
	RegisterFailHandler(Fail)
	t = test
	RunSpecs(test, "Hinting client tests")
}

const (
	orgkey = "7ZP4H5XR" // env = "local", orgId = 2023
)

func getMockProvider() (*consumer.V3HTTPMockProvider, error) {
	return consumer.NewV3Pact(consumer.MockHTTPProviderConfig{
		Consumer: "ems-device-hello",
		Provider: "ems-device-hints",
		Host:     "127.0.0.1",
		Port:     5004,
		TLS:      false,
	})
}

var _ = Describe("Hinting client tests", func() {
	hr := hints.NewHintResult(hinttype.NEW_CONFIG_DATA, hinttype.NEW_POLICY, hinttype.LIVE_QUERY)
	lastUpdateTime := time.UnixMilli(1675397106000)    // Friday, February 3, 2023 4:05:06
	lastLiveQueryTime := time.UnixMilli(1672628645000) // Monday, January 2, 2023 3:04:05
	hr.LastUpdateTime = &lastUpdateTime
	hr.LastLiveQueryTime = &lastLiveQueryTime
	// ----- Begin Tests --------------------------------------------------

	version := 17
	requestLastUpdateTime := int64(0)
	requestLastLiveQueryTime := int64(0)
	messageCode := "REQ_HELLO"

	jsonHintRequest := &hints.JsonHintRequest{
		LastUpdateTime: &requestLastUpdateTime, LastLiveQueryTime: &requestLastLiveQueryTime, OrgKey: orgkey, DeviceGroupId: 456, DeviceId: 123, DeviceType: "WINDOWS", DeploymentType: "ENDPOINT", Version: &version, MessageCode: &messageCode,
	}

	hintsClient := NewHintsClient(http.DefaultClient.Do, "http://127.0.0.1:5004/hinting/v1/hints")

	It("requests hints correctly for already seen devices", func() {
		provider, err := getMockProvider()
		Expect(err).To(BeNil())
		err = provider.AddInteraction().
			Given("a healthy service").
			UponReceiving("request for hints for a device we have seen").
			WithRequest("POST", "/hinting/v1/hints", func(b *consumer.V3RequestBuilder) {
				b.JSONBody(matchers.Map{
					"deployment_type":       matchers.Equality("ENDPOINT"),
					"device_type":           matchers.Equality("WINDOWS"),
					"device_group_id":       matchers.Equality(456),
					"device_id":             matchers.Equality(123),
					"last_live_query_time":  matchers.Equality(0),
					"last_update_time":      matchers.Equality(0),
					"message_code":          matchers.Equality("REQ_HELLO"),
					"org_key":               matchers.Equality("7ZP4H5XR"),
					"remote_rules_required": matchers.Equality(false),
					"version":               matchers.Equality(17),
				})
			}).
			WillRespondWith(200, func(b *consumer.V3ResponseBuilder) {
				b.JSONBody(matchers.Map{
					"hints":                matchers.Like(hr.Hints.ToStringArray()),
					"last_live_query_time": matchers.Like(hr.LastLiveQueryTime.UnixMilli()),
					"last_update_time":     matchers.Like(hr.LastUpdateTime.UnixMilli()),
				})
			}).
			ExecuteTest(t, func(config consumer.MockServerConfig) error {
				// Act: test our API client behaves correctly
				// Initialise the API client and point it at the Pact mock server
				config.TLSConfig.InsecureSkipVerify = true

				hintResult, err := hintsClient.GetHints(jsonHintRequest)
				// Assert: check the result
				Expect(err).To(BeNil())
				Expect(hintResult).ToNot(BeNil())
				Expect(hintResult.Hints).To(BeEquivalentTo(hr.Hints))
				Expect(hintResult.LastUpdateTime).To(BeEquivalentTo(&lastUpdateTime))
				Expect(hintResult.LastLiveQueryTime).To(BeEquivalentTo(&lastLiveQueryTime))
				return err
			})

		Expect(err).To(BeNil())
	})

	It("requests hints correctly for new devices", func() {
		provider, err := getMockProvider()
		jsonHintRequest := &hints.JsonHintRequest{
			LastUpdateTime: &requestLastUpdateTime, LastLiveQueryTime: &requestLastLiveQueryTime, OrgKey: "R76KWHRN", DeviceGroupId: 1337, DeviceId: 1337, DeviceType: "WINDOWS", DeploymentType: "ENDPOINT", Version: &version, MessageCode: &messageCode,
		}
		Expect(err).To(BeNil())
		err = provider.AddInteraction().
			Given("a healthy service").
			UponReceiving("request for hints for a new device").
			WithRequest("POST", "/hinting/v1/hints", func(b *consumer.V3RequestBuilder) {
				b.JSONBody(matchers.Map{
					"deployment_type":       matchers.Equality("ENDPOINT"),
					"device_type":           matchers.Equality("WINDOWS"),
					"device_group_id":       matchers.Equality(1337),
					"device_id":             matchers.Equality(1337),
					"last_live_query_time":  matchers.Equality(0),
					"last_update_time":      matchers.Equality(0),
					"message_code":          matchers.Equality("REQ_HELLO"),
					"org_key":               matchers.Equality("R76KWHRN"),
					"remote_rules_required": matchers.Equality(false),
					"version":               matchers.Equality(17),
				})
			}).
			WillRespondWith(200, func(b *consumer.V3ResponseBuilder) {
				b.JSONBody(matchers.Map{
					"hints": matchers.Like([]string{}),
				})
			}).
			ExecuteTest(t, func(config consumer.MockServerConfig) error {
				// Act: test our API client behaves correctly
				// Initialise the API client and point it at the Pact mock server
				config.TLSConfig.InsecureSkipVerify = true

				hintResult, err := hintsClient.GetHints(jsonHintRequest)
				// Assert: check the result
				Expect(err).To(BeNil())
				Expect(hintResult).ToNot(BeNil())
				emptyHintSet := hinttype.HintSet(0)
				Expect(*hintResult.Hints).To(BeEquivalentTo(emptyHintSet))
				Expect(hintResult.LastUpdateTime).To(BeNil())
				Expect(hintResult.LastLiveQueryTime).To(BeNil())
				return err
			})

		Expect(err).To(BeNil())
	})

	It("handles 400 errors from hinting correctly", func() {
		provider, err := getMockProvider()
		Expect(err).To(BeNil())
		err = provider.AddInteraction().
			Given("an healthy service").
			UponReceiving("invalid request for hints").
			WithRequest("POST", "/hinting/v1/hints", func(b *consumer.V3RequestBuilder) {
				b.JSONBody(matchers.Map{
					"deployment_type":       matchers.Equality("ENDPOINT"),
					"device_type":           matchers.Equality("WINDOWS"),
					"device_group_id":       matchers.Equality(456),
					"device_id":             matchers.Equality(123),
					"last_live_query_time":  matchers.Equality(0),
					"last_update_time":      matchers.Equality(0),
					"message_code":          matchers.Equality("REQ_HELLO"),
					"org_key":               matchers.Equality("NOTAREALORGKEY"),
					"remote_rules_required": matchers.Equality(false),
					"version":               matchers.Equality(17),
				})
			}).
			WillRespondWith(400).
			ExecuteTest(t, func(config consumer.MockServerConfig) error {
				jsonHintRequest.OrgKey = "NOTAREALORGKEY"
				config.TLSConfig.InsecureSkipVerify = true
				_, err := hintsClient.GetHints(jsonHintRequest)
				Expect(err).ToNot(BeNil())
				return nil
			})

		Expect(err).To(BeNil())
	})

	It("returns an error when no hint request supplied", func() {
		hintResult, err := hintsClient.GetHints(nil)
		Expect(err).ToNot(BeNil())
		Expect(hintResult).ToNot(BeNil())
	})

	It("returns an error when hint response has an error", func() {
		hrData, _ := hr.MarshalJSON()
		theClient := NewHintsClient(func(req *http.Request) (*http.Response, error) {
			return &http.Response{StatusCode: 200, Body: io.NopCloser(bytes.NewBuffer(hrData))}, errors.New("bad call")
		}, "ignored")

		hintResult, err := theClient.GetHints(jsonHintRequest)
		Expect(err).ToNot(BeNil())
		Expect(hintResult).ToNot(BeNil())
	})

	It("returns an error when hint response has a 500 code", func() {
		hrData, _ := hr.MarshalJSON()
		theClient := NewHintsClient(func(req *http.Request) (*http.Response, error) {
			return &http.Response{StatusCode: 500, Body: io.NopCloser(bytes.NewBuffer(hrData))}, nil
		}, "ignored")

		hintResult, err := theClient.GetHints(jsonHintRequest)
		Expect(err).ToNot(BeNil())
		Expect(hintResult).ToNot(BeNil())
	})

	// FUTURE: Figure out how to make a test that has an error when reading the response body

	It("will return an error if the body can't be unmarshaled", func() {
		_, err := NewHintsClient(func(req *http.Request) (*http.Response, error) {
			return &http.Response{StatusCode: 200, Body: io.NopCloser(bytes.NewBufferString("BOGUS"))}, nil
		}, "ignored").GetHints(jsonHintRequest)
		Expect(err).ToNot(BeNil())
	})
})
